/****************************************************************************
   Program:     $Id: trace.c 540 2009-03-27 17:05:09Z rbeverly $
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
   Date:        $Date: 2009-03-27 13:05:09 -0400 (Fri, 27 Mar 2009) $
   Description: Traceroute wrapper
****************************************************************************/
#include "spoof.h"

#ifndef _WIN32
#include <signal.h>

#define TRWAIT 6	// six second timeout on select() reading traceroute pipe
#define MAXWAIT 30  // maximum wait of 30 seconds 
#define NUM_TR_EXEC 5

enum tr_type {UNKNOWN, STANDARD, MISSING, PARALLEL};

int childid = 0;
pid_t pid[MAX_TRACEROUTE_CHILD];

/**
 * findtr - Locate traceroute binary on system
 *
 * @param[out]  traceroute      path to traceroute binary
 */
int findtr(char *traceroute) {
    char tr[NUM_TR_EXEC][SMALLBUF] = { 	
        "/usr/local/sbin/traceroute",
        "/usr/sbin/traceroute",
        "/sbin/traceroute",
        "/usr/bin/traceroute",
        "/usr/sbin/tracepath",
        };	
    FILE *fp;
    int ret = MISSING;
    int i;

    for (i=0;i<NUM_TR_EXEC;i++) {
        fp = fopen(tr[i], "r");
        if (fp != NULL) {
            fclose(fp);
            snprintf(traceroute, BUFSIZE, "%s", tr[i]);
            ret = STANDARD;
            break;
        }
    }
    /* check for -N flag common in Linux traceroutes */
    if (checkLinuxTraceroute(traceroute)) {
        printf("** Detected batch traceroute.\n");
        printf(">> Tracing...");
        ret = PARALLEL;
    }
    return(ret);
}

int checkLinuxTraceroute(char *tracecmd) {
    FILE *fp;
    char line[BUFSIZE];
    char buf[BUFSIZE];
    int ret = 0;

    sprintf(buf, "%s 2>&1", tracecmd);
    fp = popen(buf, "r");
    while (!feof(fp)) {
        if (fgets(line, BUFSIZE, fp) != NULL) {
            if (strstr(line, "-N") != NULL) {
                ret = 1;
                break;
            }
        }
    }
    pclose(fp);

    /* try it to be extra sure */
    if (ret) {
        ret = 0;
        sprintf(buf, "%s -n -N 1 127.0.0.1 2>&1", tracecmd);
        fp = popen(buf, "r");
        while (!feof(fp)) {
            if (fgets(line, BUFSIZE, fp) != NULL) {
                if (strstr(line, "traceroute to 127.0.0.1") != NULL) {
                    ret = 1;
                    break;
                }
            }
        }
        pclose(fp);
    }
    return (ret);
}

/* Parent process caught the ALRM, terminate (long running) child */
void catchAlrm() {
    int i;
    for (i=0;i<childid;i++) {
#if DEBUG
        printf("Terminating trace PID %d (alarm)\n", pid[i]);
#endif
        kill(pid[i], SIGKILL);
    }
}

/* Terminate (long running) child */
void catch(int child, char *reason) {
#if DEBUG
    printf("Terminating child %d trace PID %d (%s)\n", child, pid[child], reason);
#endif
    kill(pid[child], SIGKILL);
}

/**
 * runtrace - perform a traceroute to a destination 
 *
 * @param[in]   dest        destination
 * @param[out]  result      traceroute output
 *
 * Performs a traceroute to a destination.  Finds the
 * traceroute binary, starts time-limited trace.  Will
 * alarm out after TRWAIT seconds
 */
int runtrace(struct in_addr *dest, char *result, int num) {
    static enum tr_type tracetype = UNKNOWN;
    static char tracecmd[BUFSIZE];
    int bytes_read[MAX_TRACEROUTE_CHILD];
    int pipefd[MAX_TRACEROUTE_CHILD][2];
    struct timeval timeout;
    int start_time = 0;
    int lastStatus = 0;
    int maxfd = 0;
    fd_set fds;
    int offset;
    int i;

    if (tracetype == UNKNOWN)
        tracetype = findtr(tracecmd);
    if (tracetype == MISSING) 
        return -1;

    if (num > MAX_TRACEROUTE_CHILD)
        num = MAX_TRACEROUTE_CHILD;

    childid = 0;
#if DEBUG
    for (i=0;i<num;i++) 
        printf(">> Running Trace (please wait): %s to %s\n", tracecmd, 
            inet_ntoa(dest[i]));
#endif

    signal(SIGALRM, (void *) catchAlrm);
    alarm(MAXWAIT + TRWAIT + 1);
    /* signals may not be reliable on all architectures -- as backup,
       check elapsed time */
    start_time = time(NULL);
    lastStatus = time(NULL);

    /* open a pipe.  pipefd[0] is read end, pipefd[1] is write side */
    for (i=0;i<num;i++) {
        if (pipe(pipefd[i]) < 0)
            perror("pipe");
    }

    for (i=0;i<num;i++) {
        /* spawn a process to perform the trace */
        switch (pid[childid]=fork()) {
            case -1:
                return(-1);
            case 0: //child
#if DEBUG
                printf(">> Child ID: %d tracing to: %s\n", childid, inet_ntoa(dest[childid]));
#endif
                dup2(pipefd[childid][1], 2); // file descrip 2 is stderr 
                dup2(pipefd[childid][1], 1); // file descrip 1 is stdout
                if (tracetype == PARALLEL)
                    execl(tracecmd, tracecmd, "-n", "-N 1", inet_ntoa(dest[childid]), (char *) NULL);
                else
                    execl(tracecmd, tracecmd, "-n", inet_ntoa(dest[childid]), (char *) NULL);
                exit(0);
            default: //parent 
#if DEBUG
                printf("Parent spawned PID %d\n", pid[childid]);
#endif
                childid++; 
        }
    } 

    for (i=0;i<childid;i++) {
        bytes_read[i] = 0;
        if (pipefd[i][0] > maxfd)
            maxfd = pipefd[i][0];
    }

    /* while there is data to read off the pipe */
    while (TRUE) {
        if ( (time(NULL) - start_time) > MAXWAIT )  {
            for (i=0;i<childid;i++) {
                catch(i, "too long");
            }
            break;
        }
        if ( (time(NULL) - lastStatus) > 2) {
            lastStatus = time(NULL);
            printf(".");
            fflush(stdout);
        }

        /* select reset */
        FD_ZERO(&fds);
        for (i=0;i<childid;i++) {
            FD_SET(pipefd[i][0], &fds);
        }
        timeout.tv_sec = 5;
        timeout.tv_usec = 0;

        /* find a pipe with pending data from which to read */
        if (select(maxfd+1, &fds, (fd_set *) NULL, (fd_set *) NULL, &timeout) <= 0)
            break;

        /* read from pipes which have pending data */
        for (i=0;i<childid;i++) {
            if ( FD_ISSET(pipefd[i][0], &fds) ) {
                if (bytes_read[i] >= BIGBUF)
                    break;
                offset = bytes_read[i] + (BIGBUF*i);
                if (read(pipefd[i][0], &(result[offset]), 1))
                    bytes_read[i]++;
                else
                    catch(i, "unresponsive");
            }
        }
    }

    /* terminate strings with the big buffer */
    for (i=0;i<childid;i++) {
        offset = bytes_read[i] + (BIGBUF*i);
        result[offset] = '\0';
    }

    /* return the sum of read bytes across all traces */
    offset = 0;
    for (i=0;i<childid;i++) 
        offset += bytes_read[i];

    return offset;
}

#if 0
int main() {
    struct in_addr *dests;
    char *result;
    int numdests = 3;
    int i;

    result = calloc(numdests, BIGBUF);
    dests  = calloc(numdests, sizeof(struct in_addr));

    inet_aton("64.233.161.99", &dests[0]);
    inet_aton("18.7.22.83", &dests[1]);
    inet_aton("128.61.2.1", &dests[2]);

    runtrace(dests, result, numdests);
    for (i=0;i<numdests;i++) {
        printf("Trace %d: %s\n", i, &result[BIGBUF*i]);
    }
    exit(0);
}
#endif

#endif
