/****************************************************************************
   Program:     $Id: util.c 545 2009-03-28 02:36:31Z rbeverly $
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
   Date:        $Date: 2009-03-27 22:36:31 -0400 (Fri, 27 Mar 2009) $
   Description: Spoofer utility routines
****************************************************************************/
#include "spoof.h"
#include <stdarg.h>

/**
 * Fatal error condition function 
 */
void fatal(char *fmt, ...) {
    va_list ap;
    fflush(stdout);
    va_start(ap, fmt);
    fprintf(stderr, "*** ");
    vfprintf(stderr, fmt, ap);
    va_end(ap);
#ifdef _WIN32
    winpause("Press Enter to Exit.");
#endif
    exit(-1);
}

/**
 * Fatal error condition function; prints error
 */
void pfatal(char *err) {
    fflush(stdout);
    perror(err);
    exit(-1);
}

/**
 * binDump -  Debug routine for examining raw packets/buffer/char strings 
 *
 * @param[in]   msg     pointer to message buffer
 * @param[in]   len     length of buffer to print
 */
void binDump(unsigned char *msg, int len) {
    int i;
    for (i=1;i<=len;i++) {
        if (len > 16) {
            printf("%02x ", *msg++);
            if (i%4 == 0) printf("\n");
        } else {
            printf("%02x", *msg++);
        }
    }
    printf("\n");
}

/**
 * byte32Swap - A ntohl() for big endian architectures who null macro it 
 */
unsigned long byte32Swap(unsigned long in) {
    return ( (in & 0x000000ff)<<24 | (in & 0x0000ff00)<<8 |
             (in & 0x00ff0000)>>8 | (in & 0xff000000)>>24 );
}

/**
 * byte16Swap - A ntoh() for big endian architectures who null macro it 
 */
unsigned short byte16Swap(unsigned short in) {
    return ( (in & 0x00ff)<<8 | (in & 0xff00)>>8 );
}

/**
 * checkroot -  Check for root priviledges 
 */
void checkroot() {
#ifndef _WIN32
    if ( (getuid() != 0) && (geteuid() != 0) ) {
        warn("This application needs root privileges to send probes");
        warn("with spooofed source addresses (using raw sockets).");
        warn("See: http://%s/faq.php for details.", REPORT_HOST);
        fatal("Must be run as root!\n");
    }
#endif
}

/**
 * win_init -  Windows socket API startup 
 */
void win_init() {
#ifdef _WIN32
    WSADATA wsd;
    if (WSAStartup(MAKEWORD(2, 2), &wsd) != 0) {
        winfatal("** WSAStartup() failed");
    }
#endif
}

/**
 * banner - Hi there 
 */
void banner() {
    printf(">> MIT ANA Spoofer Project\n");
    printf(">> IP Source Spoofing Tester v%s\n", PACKAGE_VERSION);
    printf(">> http://%s\n", REPORT_HOST);
    printf(">> Copyright (c) 2004-2009 Robert Beverly\n");
    printf(">>\n");
}

void banner2() {
    printf(">> This tester probes your ISP:\n");
    printf(">>\t- Sends non-spoofed source validation UDP packets\n");
    printf(">>\t- Sends spoofed source probes to multiple destinations\n");
    printf(">>\t- Runs traceroute to each destination\n");
    printf(">>\t- Returns a web page with a graphical picture of results\n");
    printf(">>\n");
}

/**
 * confirm - Ask for user input before proceeding
 */
void confirm() {
    char buf[SMALLBUF];

    printf("Continue? [y/N]....");
    fflush(stdout);
    fgets(buf, SMALLBUF, stdin);
    if ( ((buf[0] != 'y') && (buf[0] != 'Y')) || (strlen(buf) != 2) ) 
        fatal("Aborting.\n");
}

/**
 * spoofSleep -  Portable microsecond random sleep 
 */
void spoofSleep() {
    unsigned int gap;

    /* random interpacket delay 0-0.5s */
    gap = (unsigned int)((double)rand() / ((double)RAND_MAX + 1) * 500);
#ifdef _WIN32
    Sleep(gap);
#else
    usleep(gap * 1000);
#endif
}

/**
 * timestamp - form time/date string
 */
char *timestamp() {
   static char str[BUFSIZE];
   time_t clock;
   struct tm *t;

   clock = time(NULL);
   t = localtime(&clock);
   snprintf(str, sizeof(str), "%02d/%02d %02d:%02d:%02d", t->tm_mon + 1,
      t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
   return (str);
}

/**
 * streamSetup - set stdout to be unbuffered, replicate stdout
 *               to the stderr file descriptor.  useful with
 *               the authorization services API (for osx gui)
 *               which doesn't pipe back stderr and prevents
 *               weird interleaving of stderr, stdout messages.
 */
void streamSetup() {
    if (setvbuf(stdout, NULL, _IONBF, 0) == EOF) {
        perror("setvbuf");
    }
    if (dup2(1,2) < 0) {
        perror("dup2");
    }
}
