/****************************************************************************
   Program:     $Id: wintrace.c 543 2009-03-28 02:23:09Z rbeverly $
   Author:      Rob Beverly <rbeverly@mit.edu>
   Date:        $Date: 2009-03-27 22:23:09 -0400 (Fri, 27 Mar 2009) $
   Description: WIN32-Specific traceroute wrapper
****************************************************************************/
#include "spoof.h"

#ifdef _WIN32

#define TRWAIT 20	// wait up to 20 seconds for results
#define TRDEBUG 0

HANDLE ghMutex;
int children_living = 0;
int alive_indicator[MAX_TRACEROUTE_CHILD];
HANDLE gStdOut_Rd[MAX_TRACEROUTE_CHILD];
HANDLE gStdOut_Wr[MAX_TRACEROUTE_CHILD];

/* thread-specific data */
typedef struct MyData {
    int id;
    struct in_addr *dest;
    char *result;
} MYDATA, *PMYDATA; 


/* thread which performs tracing work */
DWORD WINAPI traceThread(LPVOID lpParam) {
    PMYDATA pDataArray;
    char tracecmd[BUFSIZE];
    int bytes_read = 0;

    /* from: http://msdn.microsoft.com/en-us/library/ms682499.aspx */
    STARTUPINFO siStartInfo;
    PROCESS_INFORMATION piProcInfo;
    SECURITY_ATTRIBUTES saAttr;
    DWORD dwRead;
    BOOL bSuccess = FALSE;

    pDataArray = (PMYDATA) lpParam;
#if TRDEBUG
    printf(">> Running Trace to: %s [thread: %d]\n", 
        inet_ntoa(*(pDataArray->dest)), pDataArray->id);
#endif
    sprintf(tracecmd, "tracert -d %s", inet_ntoa(*(pDataArray->dest)));

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    CreatePipe(&gStdOut_Rd[pDataArray->id], &gStdOut_Wr[pDataArray->id], &saAttr, 0);
    SetHandleInformation(gStdOut_Rd[pDataArray->id], HANDLE_FLAG_INHERIT, 0);

    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
    siStartInfo.cb = sizeof(STARTUPINFO);
    siStartInfo.hStdError = NULL;
    siStartInfo.hStdOutput = gStdOut_Wr[pDataArray->id];
    siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

    CreateProcess(NULL, tracecmd, NULL, NULL, TRUE, 0,
        NULL, NULL, &siStartInfo, &piProcInfo);
    CloseHandle(gStdOut_Wr[pDataArray->id]);
    while (1) {
        if (children_living == 0) {
#if TRDEBUG
            printf("I should die [thread: %d]\n", pDataArray->id);
#endif
            ExitThread(0);
        }
        bSuccess = PeekNamedPipe(gStdOut_Rd[pDataArray->id], NULL, 0,
            NULL, &dwRead, NULL);
        if (bytes_read + dwRead >= BIGBUF)
            break;
        if (dwRead > 0) {
            bSuccess = ReadFile(gStdOut_Rd[pDataArray->id], 
                pDataArray->result + bytes_read, BUFSIZE, &dwRead, NULL);
            bytes_read += dwRead;
            if (!bSuccess || dwRead == 0) break;
        } else {
            Sleep(500);
        }
#if TRDEBUG
        printf("Thread: %d Read: [%d] Total: [%d]\n",
                pDataArray->id, dwRead, bytes_read);
#endif
    }
#if TRDEBUG
    printf("Thread %d exit.\n", pDataArray->id);
#endif
    WaitForSingleObject(ghMutex, INFINITE);
    alive_indicator[pDataArray->id] = 0; 
    children_living--;
    ReleaseMutex(ghMutex);
    ExitThread(0);
    return 0;
}

/**
 * runtrace - perform a traceroute to a destination 
 *
 * @param[in]   dest        destination
 * @param[out]  result      traceroute output
 * @param[in]   num     number of destinations
 *
 * Performs a traceroute to a destination.  Uses tracert
 * binary, starts time-limited trace.  Will terminate
 * after TRWAIT seconds
 */
int runtrace(struct in_addr *dest, char *result, int num) {
    PMYDATA pDataArray[MAX_TRACEROUTE_CHILD];
    HANDLE hThreadArray[MAX_TRACEROUTE_CHILD];
    MSG msg;
    int lastStatus = 0;
    int i;

    /* Reset timer */
    memset(&msg, 0, sizeof(msg));
    SetTimer(NULL, 0, TRWAIT * 1000, NULL);
    lastStatus = time(NULL);
    if (num > MAX_TRACEROUTE_CHILD)
        num = MAX_TRACEROUTE_CHILD;

    /* Initialize thread data */
    for (i=0;i<num;i++) {
        pDataArray[i] = (PMYDATA) HeapAlloc(GetProcessHeap(),
            HEAP_ZERO_MEMORY, sizeof(MYDATA));
        pDataArray[i]->id = i;
        pDataArray[i]->dest = dest + i;
        pDataArray[i]->result = result + (i*BIGBUF); 
        children_living++;
        alive_indicator[i] = 1;
    }
    ghMutex = CreateMutex(NULL, FALSE, NULL);

    /* Start threads */
    for (i=0;i<num;i++) {
        hThreadArray[i] = CreateThread(NULL, 0, traceThread, 
            pDataArray[i], 0, NULL);
    }
    
    /* Wait for threads to finish, or timeout */
    while (children_living > 0) {
        Sleep(1000);
        if ( (time(NULL) - lastStatus) > 2) {
            lastStatus = time(NULL);
            printf(".");
            fflush(stdout);
        } 
#if TRDEBUG
        printf("%d threads active\n", children_living);
#endif
        PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);
        if (msg.message == WM_TIMER) {
            printf("finishing.");
            WaitForSingleObject(ghMutex, INFINITE);
            children_living = 0;
            ReleaseMutex(ghMutex);
        }
    }

    /* Shut down threads */
    Sleep(1000);
    for (i=0;i<num;i++) {
#if TRDEBUG
        if (alive_indicator[i]) 
            printf("Thread %d still running. shut it down.\n", i);
#endif
        CloseHandle(hThreadArray[i]);
        if (pDataArray[i] != NULL) {
            HeapFree(GetProcessHeap(), 0, pDataArray[i]);
            pDataArray[i] = NULL;
        }
    }
    return(1);
}

#if 0
int main () {
    struct in_addr *dests;
    char *result;
    int numdests = 5;
    int i;

    result = calloc(numdests, BIGBUF);
    dests  = calloc(numdests, sizeof(struct in_addr));

    dests[0].s_addr = inet_addr("192.172.226.5");
    dests[1].s_addr = inet_addr("128.61.2.1");
    dests[2].s_addr = inet_addr("128.63.2.52");
    dests[3].s_addr = inet_addr("192.168.1.1");
    dests[4].s_addr = inet_addr("128.89.80.99");
//    dests[0].s_addr = inet_addr("64.233.161.99");
//    dests[1].s_addr = inet_addr("18.7.22.83");
//    dests[2].s_addr = inet_addr("128.61.2.1");

    runtrace(dests, result, numdests);
    for (i=0;i<numdests;i++) {
        printf("WinTrace %d: %s\n", i, &result[BIGBUF*i]);
    }
    exit(0);
}
#endif

#endif
